# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import os
import shutil
import tempfile

__all__ = ['TmpDir']

class TmpDir(object):
	_dn = None
	_rm = None

	def __init__(self, suffix = None, prefix = None, dir = None, rmtree = False):
		if prefix and not prefix.endswith('.'):
			prefix = "%s." % prefix
		self._dn = tempfile.mkdtemp(suffix = suffix, prefix = prefix, dir = dir)
		self._rm = rmtree

	def __enter__(self):
		return self._dn

	def __exit__(self, *args):
		if self._rm:
			self.rmtree()

	def __str__(self):
		return self._dn

	def rmtree(self):
		shutil.rmtree(self._dn)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.utils import sh, cmd

__all__ = ['deploy']

def deploy(env):
	for s in env.settings.sections():
		if s.startswith('docker-compose:'):
			name = s.replace('docker-compose:', '', 1).strip()
			cfg = env.settings[s]
			_composeBuild(env, name, cfg)

def _composeBuild(env, name, cfg):
	env.log("docker-compose build %s" % name)
	if cfg.getboolean('build', fallback = True):
		workdir = cfg.get('path')
		oldwdir = sh.getcwd()
		try:
			sh.chdir(workdir)
			cmd.callCheck(['docker-compose', 'build'])
		finally:
			sh.chdir(oldwdir)



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import os
import os.path
import shutil as _sh

from contextlib import contextmanager

from .tmpdir import TmpDir
from .tmpfile import TmpFile

__all__ = ['makedirs', 'chmod', 'chown', 'mktmp', 'mktmpdir', 'getcwd', 'chdir',
	'getuid', 'getgid', 'lockd']

class _ShUtil(object):
	def __init__(self):
		self.makedirs = os.makedirs
		self.chmod = os.chmod
		self.chown = _sh.chown
		self.mktmp = self._mktmp
		self.mktmpdir = self._mktmpdir
		self.getcwd = os.getcwd
		self.chdir = os.chdir
		self.getuid = os.getuid
		self.getgid = os.getgid
		self.lockd = self._lockd

	def _mktmp(self, suffix = None, prefix = None, dir = None, remove = False):
		return TmpFile(suffix = suffix, prefix = prefix, dir = dir, remove = remove)

	def _mktmpdir(self, suffix = None, prefix = None, dir = None, rmtree = False):
		return TmpDir(suffix = suffix, prefix = prefix, dir = dir, rmtree = rmtree)

	def _lockd(self, path):
		fn = os.path.join(path, '.sadmlock')
		try:
			with open(fn, 'x') as fh:
				fh.write('1\n')
				fh.flush()
			yield
		finally:
			os.unlink(fn)

shutil = _ShUtil()

def makedirs(name, mode = 0o0755, exists_ok = False):
	return shutil.makedirs(name, mode = mode, exist_ok = exists_ok)

def chmod(path, mode):
	return shutil.chmod(path, mode)

def chown(path, user = None, group = None):
	return shutil.chown(path, user = user, group = group)

def mktmp(suffix = None, prefix = None, dir = None, remove = False):
	return shutil.mktmp(suffix = suffix, prefix = prefix, dir = dir, remove = remove)

def mktmpdir(suffix = None, prefix = None, dir = None, rmtree = False):
	return shutil.mktmpdir(suffix = suffix, prefix = prefix, dir = dir, rmtree = rmtree)

def getcwd():
	return shutil.getcwd()

def chdir(path):
	return shutil.chdir(path)

def getuid():
	return shutil.getuid()

def getgid():
	return shutil.getgid()

@contextmanager
def lockd(path):
	try:
		yield shutil.lockd(path)
	finally:
		pass



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm.listen.webhook.handlers import repo

def test_hook_push(listen_wapp):
	with listen_wapp(profile = 'bitbucket') as wapp:
		wapp.POST('bitbucket_push', repo.handle, 'bitbucket', 'testing', 'push')
		assert wapp.response == 'OK'



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from bottle import HTTPError
from pytest import raises

from _sadm.listen.webhook import repo

PROVIDERS = [
	'bitbucket',
	'testing',
]

def test_providers_list():
	assert sorted(repo._provider.keys()) == PROVIDERS

def test_noprovider():
	with raises(HTTPError) as exc:
		repo.WebhookRepo('noprov', 'testing')
	err = exc.value
	assert err.status_code == 400
	assert err.body == 'webhook invalid provider: noprov'

def test_norepo(listen_wapp):
	with listen_wapp():
		with raises(HTTPError) as exc:
			repo.WebhookRepo('testing', 'norepo')
		err = exc.value
		assert err.status_code == 400
		assert err.body == 'webhook testing repo not found: norepo'



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm import log
from _sadm.listen.handlers import exec as _exec
from _sadm.listen.webhook.handlers import repo
from _sadm.listen.plugin import ResponsePlugin

__all__ = ['init']

def init(wapp, config):
	wapp.install(ResponsePlugin())

	wapp.route('/_/exec/<task>/<action>', 'POST', _exec.handle, name = '_exec')

	initDone = {}
	for sect in config.sections():
		if sect.startswith('sadm.webhook:'):
			if not initDone.get('webhook', False):
				log.debug('enable webhook')
				_initWebhooks(wapp)
				initDone['webhook'] = True

def _initWebhooks(wapp):
	wapp.route('/hook/<provider>/<name>/<action>', 'POST', repo.handle, name = 'hook.repo')



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import bottle
from configparser import ConfigParser, ExtendedInterpolation

from _sadm import libdir, log, version
from _sadm.listen import errors, handlers
from _sadm.utils import path

__all__ = ['wapp', 'config', 'init']

_wappcfg = libdir.fpath('listen', 'wapp.conf')
_cfgfn = path.join(path.sep, 'etc', 'opt', 'sadm', 'listen.cfg')

wapp = bottle.Bottle()
config = None

def _newConfig(fn):
	c = ConfigParser(
		defaults = None,
		allow_no_value = False,
		delimiters = ('=',),
		comment_prefixes = ('#',),
		strict = True,
		interpolation = ExtendedInterpolation(),
		default_section = 'default',
	)
	with open(fn, 'r') as fh:
		c.read_file(fh)
	return c

def init(cfgfn = _cfgfn):
	global config

	wapp.config.load_config(_wappcfg)
	bottle.TEMPLATE_PATH = []

	config = _newConfig(cfgfn)
	log.init(config.get('sadm', 'log', fallback = 'error'))
	log.debug(version.string('sadm-listen'))

	rmplugins = [p.strip() for p in wapp.config.get('plugins.uninstall', '').split(' ')]
	for p in rmplugins:
		if p != '':
			log.debug("plugin uninstall %s" % p)
			wapp.uninstall(p)

	errors.init(wapp)
	handlers.init(wapp, config)

	return wapp



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm import log, cfg
from _sadm.cmd import flags
from _sadm.deploy import cmd, cfgfile
from _sadm.env import Env
from _sadm.utils import sh, path
from _sadm.utils.cmd import call

def cmdArgs(parser):
	p = parser.add_parser('deploy', help = 'deploy sadm.env')
	p.set_defaults(command = 'deploy')

def main(args, sumode):
	log.debug("deploy %s sumode=%s" % (args.env, sumode))
	if sumode == 'not':
		dn = path.join('~', '.local', 'sadm', 'deploy')
		sh.makedirs(dn, mode = 0o750, exists_ok = True)
		with sh.lockd(dn):
			env = Env('deploy', args.env, cfg.new(cfgfile = cfgfile))
			for rc in (
				_sumode(env, 'pre'),
				_usermode(env),
				_sumode(env, 'post'),
			):
				if rc != 0:
					return rc
	else:
		return cmd.run(args.env, sumode)

def _sumode(env, step):
	sumode = '-'.join(['--sumode', step])
	log.debug("call sumode %s" % sumode)
	sudo = env.config.get('deploy', 'sudo.command')
	cmd = sudo.strip().split()
	cmd.extend(flags.cmdline.split())
	cmd.append(sumode)
	log.debug("sumode cmd %s" % ' '.join(cmd))
	return call(cmd)

def _usermode(env):
	return cmd.run(env, 'not')



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import bottle

from os import path

from _sadmtest import mock
from _sadmtest.listen.webhook.repo.provider import TestingProvider
from _sadmtest.wapp import TestingWebapp

import _sadm.listen.wapp
import _sadm.listen.webhook.repo
_sadm.listen.webhook.repo._provider['testing'] = TestingProvider()

class ListenWebapp(TestingWebapp):
	name = 'listen'

	def __enter__(self):
		fn = path.join('tdata', 'listen.cfg')
		_sadm.listen.wapp.wapp = _sadm.listen.wapp.init(cfgfn = fn)
		if self.profile != '':
			fn = path.join('tdata', 'listen', self.profile, 'listen.cfg')
			_sadm.listen.wapp.config.read(fn)
		with mock.utils(_sadm.listen.wapp.config):
			return self

	def __exit__(self, exc_type, exc_val, exc_tb):
		del _sadm.listen.wapp.config
		del _sadm.listen.wapp.wapp
		_sadm.listen.wapp.wapp = bottle.Bottle()



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from bottle import response

from _sadm import log

__all__ = ['ResponsePlugin']

class ResponsePlugin(object):
	api = 2
	name = 'sadm.listen.response'

	def setup(self, wapp):
		log.debug('setup')

	def apply(self, callback, ctx):
		log.debug("apply for rule: %s" % ctx.rule)
		def wrapper(*args, **kwargs):
			log.debug('apply.wrapper')
			resp = callback(*args, **kwargs)
			response.headers['Content-Type'] = 'text/plain; charset=UTF-8'
			return resp
		return wrapper



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from contextlib import contextmanager

from _sadmtest.mock.utils.cmd import MockCmdProc
from _sadmtest.mock.utils.path import MockPath
from _sadmtest.mock.utils.sh import MockShUtil

import _sadm.utils.cmd
import _sadm.utils.path
import _sadm.utils.sh

__all__ = ['utils']

def _mockUtils(cfg):
	_sadm.utils.cmd.proc = MockCmdProc(cfg)
	_sadm.utils.path._path = MockPath(cfg)
	_sadm.utils.sh.shutil = MockShUtil(cfg)

def _mockUtilsCheck():
	_sadm.utils.path._path.check()
	_sadm.utils.sh.shutil.check()
	_sadm.utils.cmd.proc.check()

def _mockUtilsRestore():
	_sadm.utils.cmd.proc = _sadm.utils.cmd._ProcMan()
	_sadm.utils.path._path = _sadm.utils.path._Path()
	_sadm.utils.sh.shutil = _sadm.utils.sh._ShUtil()

@contextmanager
def utils(cfg, tag = 'utils'):
	mockcfg = None
	sect = "_sadmtest.mock.%s" % tag
	if cfg and cfg.has_section(sect):
		mockcfg = cfg[sect]
	try:
		_mockUtils(mockcfg)
		yield
		_mockUtilsCheck()
	finally:
		_mockUtilsRestore()



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from configparser import ConfigParser
from os import path
from pytest import raises

from _sadm.listen import wapp, errors

def test_default_config():
	assert wapp.config is None

def test_config_error():
	cfgfn = path.join('tdata', 'no-listen.cfg')
	with raises(FileNotFoundError):
		w = wapp.init(cfgfn = cfgfn)

ROUTES = [
	'_exec /_/exec/<task>/<action> POST',
]

def test_default_wapp():
	cfgfn = path.join('tdata', 'listen.cfg')
	assert not errors._initDone
	w = wapp.init(cfgfn = cfgfn)
	assert errors._initDone
	assert sorted([p.name for p in w.plugins]) == ['sadm.listen.response']
	routes = []
	for r in w.routes:
		routes.append(' '.join([str(r.name), r.rule, r.method]))
	assert sorted(routes) == ROUTES

def test_wapp(listen_wapp):
	with listen_wapp() as w:
		assert w.name == 'listen'
		assert w.response is None



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import os
import tempfile

__all__ = ['TmpFile']

class TmpFile(object):
	_fd = None
	_fn = None
	_rm = None

	def __init__(self, suffix = None, prefix = None, dir = None, remove = False):
		if prefix and not prefix.endswith('.'):
			prefix = "%s." % prefix
		self._fd, self._fn = tempfile.mkstemp(suffix = suffix,
			prefix = prefix, dir = dir, text = False)
		self._rm = remove

	def __enter__(self):
		return self

	def __exit__(self, *args):
		self.close()
		if self._rm:
			self.unlink()

	def close(self):
		os.close(self._fd)

	def unlink(self):
		os.unlink(self._fn)

	def write(self, data):
		if isinstance(data, str):
			data = data.encode('utf-8')
		os.write(self._fd, data)

	def name(self):
		return self._fn



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

# ~ from _sadm.utils import systemd

__all__ = ['deploy']

# run as root at first pass
sumode = 'pre'

def deploy(env):
	pass
	# ~ systemd.restart('netfilter-persistent')



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

__all__ = ['TestingProvider']

class TestingProvider(object):

	def auth(self, req, cfg):
		pass

	def repoArgs(self, obj, cfg):
		return {}



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import sys
from _sadm import log, env
from _sadm.cmd import flags
from _sadm.cmd.deploy import loader
from _sadm.cmd.deploy import deploy
from _sadm.utils import sh

_parser = flags.new('sadm', desc = 'deploy sadm env')

def _getArgs():
	subparser = _parser.add_subparsers(title = 'commands',
		description = 'run `sadm command -h` for more information',
		help = 'description')
	loader.cmdArgs(subparser)
	deploy.cmdArgs(subparser)
	return flags.parse(_parser)

def main():
	sumode = 'not'
	if '--sumode-pre' in sys.argv:
		sys.argv.remove('--sumode-pre')
		sumode = 'pre'
	elif '--sumode-post' in sys.argv:
		sys.argv.remove('--sumode-post')
		sumode = 'post'

	args = _getArgs()
	log.debug("deploy %s/%s sumode=%s" % (args.profile, args.env, sumode))

	if sumode == 'not' and sh.getuid() == 0:
		log.error('do not run as root')
		return 9

	try:
		cmd = args.command
	except AttributeError:
		log.error('invalid usage')
		_parser.print_usage()
		return 1
	log.debug("dispatch command %s" % cmd)

	if args.command == 'import':
		return loader.main(args)

	return deploy.main(args, sumode)

if __name__ == '__main__':
	sys.exit(main())



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

from _sadm import log, env, deploy

def run(envobj, sumode):
	log.debug("***** RUN %s SUMODE %s *****" % (envobj, sumode))
	rc, _ = env.run('deploy', envobj, 'deploy', cfgfile = deploy.cfgfile, sumode = sumode)
	return rc



# Copyright (c) Jeremías Casteglione <jrmsdev@gmail.com>
# See LICENSE file.

import bottle
import json

from io import BytesIO
from os import path

class TestingWebapp(object):
	profile = ''
	name = None
	response = None

	def __init__(self, profile):
		if profile == '':
			self.profile = '.'
		else:
			self.profile = path.join(*profile.split('/'))

	def __postData(self, name):
		fn = path.join('tdata', self.name, self.profile, "%s.json" % name)
		with open(fn, 'r') as fh:
			return json.load(fh)

	def POST(self, pdata, callback, *args):
		post = self.__postData(pdata)
		# TODO: set headers from posdata
		data = json.dumps(post['data']).encode('utf-8')
		bottle.request.environ['CONTENT_LENGTH'] = str(len(data))
		bottle.request.environ['wsgi.input'] = BytesIO()
		bottle.request.environ['wsgi.input'].write(data)
		bottle.request.environ['wsgi.input'].seek(0, 0)
		resp = callback(*args)
		self.response = resp.strip()
